︠e38cfd34-664e-46d3-9ad7-da142df53dcdi︠
%html
<h1>
    A beta-splitting model for evolutionary trees
</h1>
<h2>
    2014, 2015, Raazesh Sainudiin and Amandine Veber
</h2>
<p>
    This is code accompanying the technical report 'A beta-splitting model for evolutionary trees' by Raazesh Sainudiin and Amandine Veber. The code is for clarity and simplicity of implementation (it is not efficient for large scale simulations).
</p>

︡6c6b8aae-fac8-41cd-b5c3-d964fbebfbae︡{"done":true,"html":"<h1>\n    A beta-splitting model for evolutionary trees\n</h1>\n<h2>\n    2014, 2015, Raazesh Sainudiin and Amandine Veber\n</h2>\n<p>\n    This is code accompanying the technical report 'A beta-splitting model for evolutionary trees' by Raazesh Sainudiin and Amandine Veber. The code is for clarity and simplicity of implementation (it is not efficient for large scale simulations).\n</p>"}
︠fc4ec488-789b-4454-834d-46f62708b024︠
def split01ScaledCD(x,I):
    '''x \in (0,1), c=I[0] < d=I[1]'''
    c=I[0]; d=I[1]; e=c+(d-c)*x;
    return [[c,e],[e,d]];

def CountsDictWithFirstIndex(X):
    '''convert a list X into a Dictionary of counts or frequencies with first index of each Key saved'''
    CD = {}
    for i in range(len(X)):
        x=X[i]
        if (x in CD):
            CD[x][1] = CD[x][1]+1
        else:
            CD[x] = [i,1]
    return CD

def SplittingPermutation(splitpointsequence):
    '''return the permutation of [n] given by the map from
    the sequence of n real numbers in the list
    splitpointsequence to an ordering by indices in [n]'''
    sss=sorted(splitpointsequence)
    return tuple([splitpointsequence.index(i)+1 for i in sss])

def MakePartitionAndTree(n,m,a,b):
    '''This creates m independent trees with n+1 leaves and
    alpha=a, beta=b
       n >= 1, where n+1 is the number of leaves
       m is the number of replicates
       a,b>-1, where (a+1,b+1) are the parameters of the
       beta distribution'''
    PlanarShapeSamples=[];
    PhyloShapeSamples=[];
    SplittingPermutationSamples=[];
    BetaD = RealDistribution('beta', [a+1, b+1],seed=0)
    show(BetaD.plot(xmin=0,xmax=1),figsize=[7,2]);
    print '\n';
    for reps in range(m):
        # generating i.i.d. samples from BetaD
        B=[BetaD.get_random_element() for _ in range(n)]
        #initialize
        LBT = LabelledBinaryTree
        t = LBT(None)
        SplitPoints=[B[0]]# keep order of split points
        Splits=split01ScaledCD(B[0],[0,1])
        t = t.binary_search_insert(B[0])

        #iterate
        for i in range(1,n):
            Widths=[x[1]-x[0] for x in Splits];
            W = GeneralDiscreteDistribution(Widths);
            nextSplitI=Splits[W.get_random_element()];
            Splits.remove(nextSplitI);
            NewLeaves=split01ScaledCD(B[i],[nextSplitI[0],
                                          nextSplitI[1]]);
            # find the split point between the new leaves
            RescaledG=NewLeaves[0][1];
            SplitPoints.append(RescaledG);
            # insert the new split point into the tree
            t = t.binary_search_insert(RescaledG);
            Splits.append(NewLeaves[0]);
            Splits.append(NewLeaves[1]);
        SplittingPermutationSamples.append(
                       SplittingPermutation(SplitPoints));
        sh=t.shape();
        PlanarShapeSamples.append(sh);
        PhyloShapeSamples.append(Graph(sh.to_undirected_graph
                      (with_leaves=True),immutable=True));
    return (PlanarShapeSamples,PhyloShapeSamples,SplittingPermutationSamples);
︡546c181d-e573-4908-ad16-61a2a122e73e︡
︠83d5305a-a31c-4a2c-a5bc-b3a087d638a6i︠
%html
<h2>
    Probability of tree under the generating-organizing DTDS Markov chain at various resolutions.
</h2>
︡c670cd6f-9896-4937-b8d5-baa538e5a76e︡{"html":"<h2>\n    Probability of tree under the generating-organizing DTDS Markov chain at various resolutions.\n</h2>\n"}︡
︠60ad0f38-8849-4c3b-8bd9-ab6c17724c5d︠
def splitsSequence(T):
    '''return a list of tuples (left,right) split sizes at each split node'''
    l = []
    T.post_order_traversal(lambda node:
       l.append((node[0].node_number(),node[1].node_number())))
    return l

def isIso(N):
    '''does node N of binary tree have the same left and right subtree shapes
       (are left and right subtrees of node N in tree isomorphic)'''
    L=Graph(N[0].canonical_labelling()
        .shape().to_undirected_graph(with_leaves=True),immutable=True)
    R=Graph(N[1].canonical_labelling()
        .shape().to_undirected_graph(with_leaves=True),immutable=True)
    #return (N.label(),N[0].label(),N[1].label(), L==R)
    return 1 if L==R else 0

def numIso(T):
    '''number of internal nodes that have isomorphic left and right sub-trees'''
    l = []
    T.post_order_traversal(lambda node: l.append(isIso(node)))
    return sum(l)

def prob_RPT(T,a,b):
    '''probability of ranked planar tree T under beta-splitting model
       a,b>-1, where (a+1,b+1) are the parameters of the beta distribution'''
    ncspS=filter(lambda x: x!=(0,0), splitsSequence(T)) # non-cherry splits
    return prod(map(lambda x: beta(x[0]+a+1,x[1]+b+1)/beta(a+1,b+1),ncspS))

def prob_PT(T,a,b):
    '''probability of planar tree T under beta-splitting model
       a,b>-1, where (a+1,b+1) are the parameters of the
       beta distribution'''
    ncspS=filter(lambda x: x!=(0,0), splitsSequence(T)) # non-cherry splits
    return prod(map(lambda x: binomial(x[0]+x[1],x[1])*beta(x[0]+a+1,x[1]+b+1)/beta(a+1,b+1),ncspS))

def prob_RT(T,a,b):
    '''probability of ranked (nonplar) tree T under beta-splitting model
       a,b>-1, where (a+1,b+1) are the parameters of the beta distribution'''
    spS=splitsSequence(T)
    numSplits=len(spS)
    ncspS=filter(lambda x: x!=(0,0), spS) # non-cherry splits
    numCherries=numSplits-len(ncspS)
    probRPT = prod(map(lambda x: beta(x[0]+a+1,x[1]+b+1)/beta(a+1,b+1), ncspS))
    return 2^(numSplits-numCherries)*probRPT

def prob_T(T,a,b):
    '''probability of tree T (phylo tree shape) under beta-splitting model
       a,b>-1, where (a+1,b+1) are the parameters of the beta distribution'''
    spS=splitsSequence(T)
    numSplits=len(spS)
    ncspS=filter(lambda x: x!=(0,0), spS) # non-cherry splits
    probPT = prod(map(lambda x: binomial(x[0]+x[1],x[1])*beta(x[0]+a+1,x[1]+b+1)/beta(a+1,b+1),ncspS))
    numIsoSplits=numIso(T)
    return 2^(numSplits-numIsoSplits)*probPT

def stats_probs_Tree(T,a,b):
    '''probability of various resolutions of tree T under beta-splitting model
       a,b>-1, where (a+1,b+1) are the parameters of the beta distribution'''
    spS=splitsSequence(T)
    numSplits=len(spS)
    ncspS=filter(lambda x: x!=(0,0), spS) # non-cherry splits
    numCherries=numSplits-len(ncspS)
    probRPT = prod(map(lambda x: beta(x[0]+a+1,x[1]+b+1)/beta(a+1,b+1), ncspS))
    catCoeff = prod(map(lambda x: binomial(x[0]+x[1],x[1]),ncspS))
    probPT = catCoeff * probRPT # prob of (non-ranked) planar tree
    probRT = 2^(numSplits-numCherries)*probRPT
    numIsoSplits=numIso(T)
    probT=2^(numSplits-numIsoSplits)*probPT
    return (numSplits,numIsoSplits,numCherries,catCoeff,probRPT,probPT,probRT,probT)
︡acc31447-b63c-41a1-9a41-cac3d69f46f3︡
︠98f01f7c-b990-40c6-82b2-3a8febf4a34ei︠
%html
<h2>
    Demonstration with a=b=0 (Yule model)
</h2>
︡e7cfdfb3-03a5-412b-80a3-a1e325815da9︡{"html":"<h2>\n    Demonstration with a=b=0 (Yule model)\n</h2>\n"}︡
︠68199269-f092-40ed-9513-197593c1f682︠
#Yule works B=0
a=0; b=0; m=10000; (bts,pts,sps)=MakePartitionAndTree(3,m,a,b)
︡f183751f-790d-4c81-ba39-e6d853aec24d︡{"once":false,"file":{"show":true,"uuid":"3bd80abb-275e-41d3-96f4-fcb752c05660","filename":"/projects/2c5f7f68-e689-4c70-a4b4-5b5d4dc4f93f/.sage/temp/compute1-us/31146/tmp_BrFgHP.svg"}}︡{"html":"<div align='center'></div>"}︡{"stdout":"\n"}︡{"stdout":"\n"}︡
︠19b25079-ee90-4b98-be0c-80dcc7884d2f︠
BtcCnts=CountsDictWithFirstIndex(sps)
for x in BtcCnts:
    print (sps[BtcCnts[x][0]],prob_RPT(bts[BtcCnts[x][0]],a,b).N(digits=4),(BtcCnts[x][1]/m).N(digits=4))
︡2a267026-500d-40a6-86f5-ea283866db60︡{"stdout":"((1, 3, 2), 0.1667, 0.1700)\n((3, 2, 1), 0.1667, 0.1666)\n((2, 1, 3), 0.1667, 0.1625)\n((3, 1, 2), 0.1667, 0.1664)\n((1, 2, 3), 0.1667, 0.1683)\n((2, 3, 1), 0.1667, 0.1662)\n"}︡
︠fb8ceec7-e6f2-4f0b-851f-50e4da0d611d︠
BtcCnts=CountsDictWithFirstIndex(bts)
for x in BtcCnts:
    print (bts[BtcCnts[x][0]],prob_PT(bts[BtcCnts[x][0]],a,b).N(digits=5),(BtcCnts[x][1]/m).N(digits=5))
︡72822b3c-7780-4181-b99e-0d5cb0b4182e︡{"stdout":"([., [[., .], .]], 0.16667, 0.17000)\n([., [., [., .]]], 0.16667, 0.16830)\n([[[., .], .], .], 0.16667, 0.16660)\n([[., [., .]], .], 0.16667, 0.16620)\n([[., .], [., .]], 0.33333, 0.32890)\n"}︡
︠c1ddaa5e-6144-4e28-99bf-c653a56ecdec︠
BtcCnts=CountsDictWithFirstIndex(pts)
for x in BtcCnts:
    pts[BtcCnts[x][0]].show()
    print (prob_T(bts[BtcCnts[x][0]],a,b).N(digits=5),(BtcCnts[x][1]/m).N(digits=5))
︡ea9f280e-82a3-4892-b9da-2ec2f30c3811︡{"once":false,"file":{"show":true,"uuid":"ea45fdf9-be3c-4cd6-a1cf-ddd2ceb10fbb","filename":"/projects/2c5f7f68-e689-4c70-a4b4-5b5d4dc4f93f/.sage/temp/compute1-us/31146/tmp_q140ST.svg"}}︡{"html":"<div align='center'></div>"}︡{"stdout":"(0.66667, 0.67110)"}︡{"stdout":"\n"}︡{"once":false,"file":{"show":true,"uuid":"62cee64e-a919-40c5-8bd4-d69b7741a1e3","filename":"/projects/2c5f7f68-e689-4c70-a4b4-5b5d4dc4f93f/.sage/temp/compute1-us/31146/tmp_9xfdFr.svg"}}︡{"html":"<div align='center'></div>"}︡{"stdout":"(0.33333, 0.32890)"}︡{"stdout":"\n"}︡
︠2aaa4f9a-5ae5-426c-88fb-d28f37b75966i︠
%html
<h2>
    Demonstration for more balanced trees with $a=b=1000 \qquad (\to \infty)$
</h2>
︡49e2542e-3d76-44b0-857d-b22f035c3b09︡{"html":"<h2>\n    Demonstration for more balanced trees with $a=b=1000 \\qquad (\\to \\infty)$\n</h2>\n"}︡
︠a59f7495-a47d-4dac-8123-4bc8c4bdd695︠
#more balanced works B -> 100
a=1000; b=1000; m=10000; (bts,pts,sps)=MakePartitionAndTree(3,m,a,b)
︡c28ff4b3-cde0-4dc7-a5fc-af5bb46a0c7d︡{"once":false,"file":{"show":true,"uuid":"9aca7c5f-b93f-42e8-9399-f451f5cf9ca4","filename":"/projects/2c5f7f68-e689-4c70-a4b4-5b5d4dc4f93f/.sage/temp/compute1-us/31146/tmp_QU5_Ko.svg"}}︡{"html":"<div align='center'></div>"}︡{"stdout":"\n"}︡{"stdout":"\n"}︡
︠6e0f5a13-4142-4166-8e63-31a6fec0581b︠
BtcCnts=CountsDictWithFirstIndex(sps)
for x in BtcCnts:
 print (sps[BtcCnts[x][0]],prob_RPT(bts[BtcCnts[x][0]],a,b).N(digits=4),(BtcCnts[x][1]/m).N(digits=4))
︡76c6db7f-0792-4a14-9a92-c6918ae5dd40︡{"stdout":"((1, 3, 2), 0.1251, 0.1229)\n((3, 2, 1), 0.1251, 0.1242)\n((2, 3, 1), 0.1251, 0.1230)\n((3, 1, 2), 0.2499, 0.2571)\n((1, 2, 3), 0.1251, 0.1216)\n((2, 1, 3), 0.2499, 0.2512)\n"}︡
︠c8c59d74-7dee-4823-8789-5b1adce5e2d4︠
BtcCnts=CountsDictWithFirstIndex(bts)
for x in BtcCnts:
    print (bts[BtcCnts[x][0]],prob_PT(bts[BtcCnts[x][0]],a,b).N(digits=4),(BtcCnts[x][1]/m).N(digits=4))
︡41629ee2-28c0-47cb-b6d2-fcd61cc4adfe︡{"stdout":"([[., [., .]], .], 0.1251, 0.1230)\n([., [., [., .]]], 0.1251, 0.1216)\n([[[., .], .], .], 0.1251, 0.1242)\n([., [[., .], .]], 0.1251, 0.1229)\n([[., .], [., .]], 0.4998, 0.5083)\n"}︡
︠7a212c0e-a1a5-4057-bb2c-78cb84830b13︠
BtcCnts=CountsDictWithFirstIndex(pts)
for x in BtcCnts:
    pts[BtcCnts[x][0]].show()
    print (prob_T(bts[BtcCnts[x][0]],a,b).N(digits=5),(BtcCnts[x][1]/m).N(digits=5))
︡bf0a8989-7d30-422e-8a10-3245dcc91bd1︡{"once":false,"file":{"show":true,"uuid":"cb1a9d37-9ef8-4187-beed-b6b181c780a3","filename":"/projects/2c5f7f68-e689-4c70-a4b4-5b5d4dc4f93f/.sage/temp/compute1-us/31146/tmp_NcCnHI.svg"}}︡{"html":"<div align='center'></div>"}︡{"stdout":"(0.50025, 0.49170)"}︡{"stdout":"\n"}︡{"once":false,"file":{"show":true,"uuid":"ff8914d1-7725-4997-9be8-01eb02d8845d","filename":"/projects/2c5f7f68-e689-4c70-a4b4-5b5d4dc4f93f/.sage/temp/compute1-us/31146/tmp_fsPWGl.svg"}}︡{"html":"<div align='center'></div>"}︡{"stdout":"(0.49975, 0.50830)"}︡{"stdout":"\n"}︡
︠ebe18014-f2fb-4102-b5dd-dfda843dc092︠
#more balanced works B -> 100
a=1000; b=1000; m=10000; (bts,pts,sps)=MakePartitionAndTree(5,m,a,b)
︡f42fa96e-728c-45c1-a5ea-16a4fd52aad4︡{"once":false,"file":{"show":true,"uuid":"1186b60e-4f11-4403-904f-4a138c0abb59","filename":"/projects/2c5f7f68-e689-4c70-a4b4-5b5d4dc4f93f/.sage/temp/compute1-us/31146/tmp_ICQjBh.svg"}}︡{"html":"<div align='center'></div>"}︡{"stdout":"\n"}︡{"stdout":"\n"}︡
︠01cbbd26-a2c0-4c6c-af0e-265ca8107481︠
BtcCnts=CountsDictWithFirstIndex(pts)
for x in BtcCnts:
    pts[BtcCnts[x][0]].show()
    print (prob_T(bts[BtcCnts[x][0]],a,b).N(digits=5),(BtcCnts[x][1]/m).N(digits=5))
︡c01b94b0-6a3a-4d36-a0c8-7074f9020588︡{"once":false,"file":{"show":true,"uuid":"db11b8dd-d4c0-4001-a0dc-50d50e7ad175","filename":"/projects/2c5f7f68-e689-4c70-a4b4-5b5d4dc4f93f/.sage/temp/compute1-us/31146/tmp_pU4sJn.svg"}}︡{"html":"<div align='center'></div>"}︡{"stdout":"(0.093984, 0.098800)"}︡{"stdout":"\n"}︡{"once":false,"file":{"show":true,"uuid":"7511f645-d9ae-4dba-a9d6-29df408ce1e1","filename":"/projects/2c5f7f68-e689-4c70-a4b4-5b5d4dc4f93f/.sage/temp/compute1-us/31146/tmp_LcO9ZO.svg"}}︡{"html":"<div align='center'></div>"}︡{"stdout":"(0.37463, 0.37940)"}︡{"stdout":"\n"}︡{"once":false,"file":{"show":true,"uuid":"b080cac5-2a56-4b23-9507-3c7f6dd8f27f","filename":"/projects/2c5f7f68-e689-4c70-a4b4-5b5d4dc4f93f/.sage/temp/compute1-us/31146/tmp_Wj2xDa.svg"}}︡{"html":"<div align='center'></div>"}︡{"stdout":"(0.015703, 0.014600)"}︡{"stdout":"\n"}︡{"once":false,"file":{"show":true,"uuid":"94e1e837-ecd1-4a6d-ab31-eba59b9ad4d7","filename":"/projects/2c5f7f68-e689-4c70-a4b4-5b5d4dc4f93f/.sage/temp/compute1-us/31146/tmp_NGx1zX.svg"}}︡{"html":"<div align='center'></div>"}︡{"stdout":"(0.015687, 0.016600)"}︡{"stdout":"\n"}︡{"once":false,"file":{"show":true,"uuid":"633e023d-a8b7-45f4-88fd-95a7edf7c51d","filename":"/projects/2c5f7f68-e689-4c70-a4b4-5b5d4dc4f93f/.sage/temp/compute1-us/31146/tmp_wJ97Ny.svg"}}︡{"html":"<div align='center'></div>"}︡{"stdout":"(0.24988, 0.24850)"}︡{"stdout":"\n"}︡{"once":false,"file":{"show":true,"uuid":"30216aba-51ed-4811-acd5-850632ffb240","filename":"/projects/2c5f7f68-e689-4c70-a4b4-5b5d4dc4f93f/.sage/temp/compute1-us/31146/tmp_RYEGgJ.svg"}}︡{"html":"<div align='center'></div>"}︡{"stdout":"(0.25012, 0.24210)"}︡{"stdout":"\n"}︡
︠6720a0c7-e4a5-44fb-a55d-05f27c3abb74︠
︠fd53ce0b-4609-4e43-bb01-37dc590e09e1︠









